@@ -2,14 +2,13 @@
 
   struct.c -
 
-  $Author: glass $
+  $Author: akr $
   created at: Tue Mar 22 18:44:30 JST 1995
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
 
 **********************************************************************/
 
-#include "ruby/ruby.h"
 #include "internal.h"
 
 VALUE rb_cStruct;
@@ -29,12 +28,6 @@ struct_ivar_get(VALUE c, ID id)
     }
 }
 
-VALUE
-rb_struct_iv_get(VALUE c, const char *name)
-{
-    return struct_ivar_get(c, rb_intern(name));
-}
-
 VALUE
 rb_struct_s_members(VALUE klass)
 {
@@ -86,6 +79,13 @@ rb_struct_members_m(VALUE obj)
     return rb_struct_s_members_m(rb_obj_class(obj));
 }
 
+NORETURN(static void not_a_member(ID id));
+static void
+not_a_member(ID id)
+{
+    rb_name_error(id, "`%"PRIsVALUE"' is not a struct member", QUOTE_ID(id));
+}
+
 VALUE
 rb_struct_getmember(VALUE obj, ID id)
 {
@@ -100,7 +100,7 @@ rb_struct_getmember(VALUE obj, ID id)
 	    return RSTRUCT_GET(obj, i);
 	}
     }
-    rb_name_error(id, "%s is not struct member", rb_id2name(id));
+    not_a_member(id);
 
     UNREACHABLE;
 }
@@ -147,20 +147,24 @@ rb_struct_modify(VALUE s)
 static VALUE
 rb_struct_set(VALUE obj, VALUE val)
 {
-    VALUE members, slot;
+    VALUE members, slot, fsym;
     long i, len;
+    ID fid = rb_frame_this_func();
 
     members = rb_struct_members(obj);
     len = RARRAY_LEN(members);
     rb_struct_modify(obj);
+    fid = rb_id_attrget(fid);
+    if (!fid) not_a_member(rb_frame_this_func());
+    fsym = ID2SYM(fid);
     for (i=0; i<len; i++) {
 	slot = RARRAY_AREF(members, i);
-	if (rb_id_attrset(SYM2ID(slot)) == rb_frame_this_func()) {
-	    return RSTRUCT_SET(obj, i, val);
+	if (slot == fsym) {
+	    RSTRUCT_SET(obj, i, val);
+	    return val;
 	}
     }
-    rb_name_error(rb_frame_this_func(), "`%s' is not a struct member",
-		  rb_id2name(rb_frame_this_func()));
+    not_a_member(fid);
 
     UNREACHABLE;
 }
@@ -485,7 +489,7 @@ struct_alloc(VALUE klass)
 VALUE
 rb_struct_alloc(VALUE klass, VALUE values)
 {
-    return rb_class_new_instance(RARRAY_LENINT(values), RARRAY_PTR(values), klass);
+    return rb_class_new_instance(RARRAY_LENINT(values), RARRAY_CONST_PTR(values), klass);
 }
 
 VALUE
@@ -596,7 +600,7 @@ rb_struct_each_pair(VALUE s)
 static VALUE
 inspect_struct(VALUE s, VALUE dummy, int recur)
 {
-    VALUE cname = rb_class_name(rb_obj_class(s));
+    VALUE cname = rb_class_path(rb_obj_class(s));
     VALUE members, str = rb_str_new2("#<struct ");
     long i, len;
     char first = RSTRING_PTR(cname)[0];
@@ -713,17 +717,17 @@ rb_struct_init_copy(VALUE copy, VALUE s)
 }
 
 static VALUE
-rb_struct_aref_id(VALUE s, ID id)
+rb_struct_aref_sym(VALUE s, VALUE name)
 {
     VALUE members = rb_struct_members(s);
     long i, len = RARRAY_LEN(members);
 
     for (i=0; i<len; i++) {
-	if (SYM2ID(RARRAY_AREF(members, i)) == id) {
+	if (RARRAY_AREF(members, i) == name) {
 	    return RSTRUCT_GET(s, i);
 	}
     }
-    rb_name_error(id, "no member '%s' in struct", rb_id2name(id));
+    rb_name_error_str(name, "no member '% "PRIsVALUE"' in struct", name);
 
     UNREACHABLE;
 }
@@ -751,7 +755,7 @@ rb_struct_aref(VALUE s, VALUE idx)
     long i;
 
     if (RB_TYPE_P(idx, T_SYMBOL)) {
-	return rb_struct_aref_id(s, SYM2ID(idx));
+	return rb_struct_aref_sym(s, idx);
     }
     else if (RB_TYPE_P(idx, T_STRING)) {
 	ID id = rb_check_id(&idx);
@@ -759,7 +763,7 @@ rb_struct_aref(VALUE s, VALUE idx)
 	    rb_name_error_str(idx, "no member '%"PRIsVALUE"' in struct",
 			      QUOTE(idx));
 	}
-	return rb_struct_aref_id(s, id);
+	return rb_struct_aref_sym(s, ID2SYM(id));
     }
 
     i = NUM2LONG(idx);
@@ -774,7 +778,7 @@ rb_struct_aref(VALUE s, VALUE idx)
 }
 
 static VALUE
-rb_struct_aset_id(VALUE s, ID id, VALUE val)
+rb_struct_aset_sym(VALUE s, VALUE name, VALUE val)
 {
     VALUE members = rb_struct_members(s);
     long i, len = RARRAY_LEN(members);
@@ -785,13 +789,13 @@ rb_struct_aset_id(VALUE s, ID id, VALUE val)
     }
 
     for (i=0; i<len; i++) {
-	if (SYM2ID(RARRAY_AREF(members, i)) == id) {
+	if (RARRAY_AREF(members, i) == name) {
 	    rb_struct_modify(s);
 	    RSTRUCT_SET(s, i, val);
 	    return val;
 	}
     }
-    rb_name_error(id, "no member '%s' in struct", rb_id2name(id));
+    rb_name_error_str(name, "no member '% "PRIsVALUE"' in struct", name);
 
     UNREACHABLE;
 }
@@ -821,7 +825,7 @@ rb_struct_aset(VALUE s, VALUE idx, VALUE val)
     long i;
 
     if (RB_TYPE_P(idx, T_SYMBOL)) {
-	return rb_struct_aset_id(s, SYM2ID(idx), val);
+	return rb_struct_aset_sym(s, idx, val);
     }
     if (RB_TYPE_P(idx, T_STRING)) {
 	ID id = rb_check_id(&idx);
@@ -829,7 +833,7 @@ rb_struct_aset(VALUE s, VALUE idx, VALUE val)
 	    rb_name_error_str(idx, "no member '%"PRIsVALUE"' in struct",
 			      QUOTE(idx));
 	}
-	return rb_struct_aset_id(s, id, val);
+	return rb_struct_aset_sym(s, ID2SYM(id), val);
     }
 
     i = NUM2LONG(idx);
@@ -949,8 +953,17 @@ rb_struct_equal(VALUE s, VALUE s2)
     return rb_exec_recursive_paired(recursive_equal, s, s2, s2);
 }
 
+/*
+ * call-seq:
+ *   struct.hash   -> fixnum
+ *
+ * Returns a hash value based on this struct's contents (see Object#hash).
+ *
+ * See also Object#hash.
+ */
+
 static VALUE
-recursive_hash(VALUE s, VALUE dummy, int recur)
+rb_struct_hash(VALUE s)
 {
     long i, len;
     st_index_t h;
@@ -958,31 +971,16 @@ recursive_hash(VALUE s, VALUE dummy, int recur)
     const VALUE *ptr;
 
     h = rb_hash_start(rb_hash(rb_obj_class(s)));
-    if (!recur) {
-	ptr = RSTRUCT_CONST_PTR(s);
-	len = RSTRUCT_LEN(s);
-	for (i = 0; i < len; i++) {
-	    n = rb_hash(ptr[i]);
-	    h = rb_hash_uint(h, NUM2LONG(n));
-	}
+    ptr = RSTRUCT_CONST_PTR(s);
+    len = RSTRUCT_LEN(s);
+    for (i = 0; i < len; i++) {
+	n = rb_hash(ptr[i]);
+	h = rb_hash_uint(h, NUM2LONG(n));
     }
     h = rb_hash_end(h);
     return INT2FIX(h);
 }
 
-/*
- * call-seq:
- *   struct.hash   -> fixnum
- *
- * Returns a hash value based on this struct's contents (see Object#hash).
- */
-
-static VALUE
-rb_struct_hash(VALUE s)
-{
-    return rb_exec_recursive_outer(recursive_hash, s, 0);
-}
-
 static VALUE
 recursive_eql(VALUE s, VALUE s2, int recur)
 {
@@ -1065,7 +1063,7 @@ rb_struct_size(VALUE s)
  *  Symbol (<code>:name</code>).
  */
 void
-Init_Struct(void)
+InitVM_Struct(void)
 {
     rb_cStruct = rb_define_class("Struct", rb_cObject);
     rb_include_module(rb_cStruct, rb_mEnumerable);
@@ -1096,5 +1094,13 @@ Init_Struct(void)
     rb_define_method(rb_cStruct, "values_at", rb_struct_values_at, -1);
 
     rb_define_method(rb_cStruct, "members", rb_struct_members_m, 0);
+}
+
+#undef rb_intern
+void
+Init_Struct(void)
+{
     id_members = rb_intern("__members__");
+
+    InitVM(Struct);
 }
