@@ -1,7 +1,6 @@
 # -*- coding: us-ascii -*-
 require 'test/unit'
 require 'thread'
-require_relative 'envutil'
 
 class TestThread < Test::Unit::TestCase
   class Thread < ::Thread
@@ -27,6 +26,13 @@ class TestThread < Test::Unit::TestCase
     end
   end
 
+  def test_inspect
+    th = Module.new {break module_eval("class C\u{30b9 30ec 30c3 30c9} < Thread; self; end")}.start{}
+    assert_match(/::C\u{30b9 30ec 30c3 30c9}:/, th.inspect)
+  ensure
+    th.join
+  end
+
   def test_main_thread_variable_in_enumerator
     assert_equal Thread.main, Thread.current
 
@@ -64,14 +70,14 @@ class TestThread < Test::Unit::TestCase
   end
 
   def test_thread_variable?
-    refute Thread.new { Thread.current.thread_variable?("foo") }.join.value
+    Thread.new { assert_not_send([Thread.current, :thread_variable?, "foo"]) }.value
     t = Thread.new {
       Thread.current.thread_variable_set("foo", "bar")
     }.join
 
-    assert t.thread_variable?("foo")
-    assert t.thread_variable?(:foo)
-    refute t.thread_variable?(:bar)
+    assert_send([t, :thread_variable?, "foo"])
+    assert_send([t, :thread_variable?, :foo])
+    assert_not_send([t, :thread_variable?, :bar])
   end
 
   def test_thread_variable_strings_and_symbols_are_the_same_key
@@ -117,7 +123,6 @@ class TestThread < Test::Unit::TestCase
     dir = File.dirname(__FILE__)
     lbtest = File.join(dir, "lbtest.rb")
     $:.unshift File.join(File.dirname(dir), 'ruby')
-    require 'envutil'
     $:.shift
     3.times {
       `#{EnvUtil.rubybin} #{lbtest}`
@@ -127,17 +132,21 @@ class TestThread < Test::Unit::TestCase
 
   def test_priority
     c1 = c2 = 0
-    t1 = Thread.new { loop { c1 += 1 } }
+    run = true
+    t1 = Thread.new { c1 += 1 while run }
     t1.priority = 3
-    t2 = Thread.new { loop { c2 += 1 } }
+    t2 = Thread.new { c2 += 1 while run }
     t2.priority = -3
     assert_equal(3, t1.priority)
     assert_equal(-3, t2.priority)
     sleep 0.5
     5.times do
+      assert_not_predicate(t1, :stop?)
+      assert_not_predicate(t2, :stop?)
       break if c1 > c2
       sleep 0.1
     end
+    run = false
     t1.kill
     t2.kill
     assert_operator(c1, :>, c2, "[ruby-dev:33124]") # not guaranteed
@@ -165,22 +174,27 @@ class TestThread < Test::Unit::TestCase
 
   def test_join
     t = Thread.new { sleep }
-    assert_nil(t.join(0.5))
+    assert_nil(t.join(0.05))
 
   ensure
     t.kill if t
   end
 
   def test_join2
-    t1 = Thread.new { sleep(1.5) }
+    ok = false
+    t1 = Thread.new { ok = true; sleep }
+    Thread.pass until ok
+    Thread.pass until t1.stop?
     t2 = Thread.new do
-      t1.join(1)
+      Thread.pass while ok
+      t1.join(0.01)
     end
     t3 = Thread.new do
-      sleep 0.5
+      ok = false
       t1.join
     end
     assert_nil(t2.value)
+    t1.wakeup
     assert_equal(t1, t3.value)
 
   ensure
@@ -232,10 +246,10 @@ class TestThread < Test::Unit::TestCase
       Thread.stop
       s += 1
     end
-    sleep 0.5
+    Thread.pass until t.stop?
     assert_equal(1, s)
     t.wakeup
-    sleep 0.5
+    Thread.pass while t.alive?
     assert_equal(2, s)
     assert_raise(ThreadError) { t.wakeup }
 
@@ -258,7 +272,7 @@ class TestThread < Test::Unit::TestCase
     assert_in_out_err([], <<-INPUT) do |r, e|
       t1 = Thread.new { sleep }
       Thread.pass
-      t2 = Thread.new { loop { } }
+      t2 = Thread.new { loop { Thread.pass } }
       Thread.new { }.join
       p [Thread.current, t1, t2].map{|t| t.object_id }.sort
       p Thread.list.map{|t| t.object_id }.sort
@@ -279,8 +293,8 @@ class TestThread < Test::Unit::TestCase
     assert_in_out_err([], <<-INPUT, %w(false 1), [])
       p Thread.abort_on_exception
       begin
-        Thread.new { raise }
-        sleep 0.5
+        t = Thread.new { raise }
+        Thread.pass until t.stop?
         p 1
       rescue
         p 2
@@ -302,8 +316,8 @@ class TestThread < Test::Unit::TestCase
     assert_in_out_err(%w(--disable-gems -d), <<-INPUT, %w(false 2), %r".+")
       p Thread.abort_on_exception
       begin
-        Thread.new { raise }
-        sleep 0.5
+        t = Thread.new { raise }
+        Thread.pass until t.stop?
         p 1
       rescue
         p 2
@@ -313,9 +327,11 @@ class TestThread < Test::Unit::TestCase
     assert_in_out_err([], <<-INPUT, %w(false true 2), [])
       p Thread.abort_on_exception
       begin
-        t = Thread.new { sleep 0.5; raise }
+        ok = false
+        t = Thread.new { Thread.pass until ok; raise }
         t.abort_on_exception = true
         p t.abort_on_exception
+        ok = 1
         sleep 1
         p 1
       rescue
@@ -329,7 +345,7 @@ class TestThread < Test::Unit::TestCase
     b = Thread.new { Thread.stop }
     c = Thread.new { Thread.exit }
     e = Thread.current
-    sleep 0.5
+    Thread.pass while a.alive? or !b.stop? or c.alive?
 
     assert_equal(nil, a.status)
     assert_predicate(a, :stop?)
@@ -351,9 +367,34 @@ class TestThread < Test::Unit::TestCase
     c.kill if c
   end
 
+  def test_switch_while_busy_loop
+    bug1402 = "[ruby-dev:38319] [Bug #1402]"
+    flag = true
+    th = Thread.current
+    waiter = Thread.start {
+      sleep 0.1
+      flag = false
+      sleep 1
+      th.raise(bug1402)
+    }
+    assert_nothing_raised(RuntimeError, bug1402) do
+      nil while flag
+    end
+    assert(!flag, bug1402)
+  ensure
+    waiter.kill.join
+  end
+
   def test_safe_level
-    t = Thread.new { $SAFE = 3; sleep }
-    sleep 0.5
+    ok = false
+    t = Thread.new do
+      EnvUtil.suppress_warning do
+        $SAFE = 3
+      end
+      ok = true
+      sleep
+    end
+    Thread.pass until ok
     assert_equal(0, Thread.current.safe_level)
     assert_equal(3, t.safe_level)
 
@@ -392,11 +433,12 @@ class TestThread < Test::Unit::TestCase
   end
 
   def test_select_wait
-    assert_nil(IO.select(nil, nil, nil, 1))
+    assert_nil(IO.select(nil, nil, nil, 0.001))
     t = Thread.new do
       IO.select(nil, nil, nil, nil)
     end
-    sleep 0.5
+    Thread.pass until t.stop?
+    assert_predicate(t, :alive?)
     t.kill
   end
 
@@ -418,7 +460,7 @@ class TestThread < Test::Unit::TestCase
       m.lock
       :foo
     end
-    sleep 0.5
+    Thread.pass until t.stop?
     t.kill
     assert_nil(t.value)
   end
@@ -492,19 +534,20 @@ class TestThread < Test::Unit::TestCase
   def test_no_valid_cfp
     skip 'with win32ole, cannot run this testcase because win32ole redefines Thread#intialize' if defined?(WIN32OLE)
     bug5083 = '[ruby-dev:44208]'
-    assert_equal([], Thread.new(&Module.method(:nesting)).value)
-    assert_instance_of(Thread, Thread.new(:to_s, &Class.new.method(:undef_method)).join)
+    assert_equal([], Thread.new(&Module.method(:nesting)).value, bug5083)
+    assert_instance_of(Thread, Thread.new(:to_s, &Class.new.method(:undef_method)).join, bug5083)
   end
 
   def make_handle_interrupt_test_thread1 flag
     r = []
     ready_p = false
+    done = false
     th = Thread.new{
       begin
         Thread.handle_interrupt(RuntimeError => flag){
           begin
             ready_p = true
-            sleep 0.5
+            sleep 0.01 until done
           rescue
             r << :c1
           end
@@ -516,6 +559,7 @@ class TestThread < Test::Unit::TestCase
     Thread.pass until ready_p
     th.raise
     begin
+      done = true
       th.join
     rescue
       r << :c3
@@ -609,7 +653,7 @@ class TestThread < Test::Unit::TestCase
         }
       }
 
-      sleep 0.1
+      Thread.pass while t.stop?
       t.raise RuntimeError
       th_waiting = false
       t.join rescue nil
@@ -619,10 +663,11 @@ class TestThread < Test::Unit::TestCase
 
   def test_handle_interrupt_and_p
     assert_in_out_err([], <<-INPUT, %w(:ok :ok), [])
-      th_waiting = true
+      th_waiting = false
 
       t = Thread.new {
         Thread.handle_interrupt(RuntimeError => :on_blocking) {
+          th_waiting = true
           nil while th_waiting
           # p shouldn't provide interruptible point
           p :ok
@@ -630,7 +675,7 @@ class TestThread < Test::Unit::TestCase
         }
       }
 
-      sleep 0.1
+      Thread.pass until th_waiting
       t.raise RuntimeError
       th_waiting = false
       t.join rescue nil
@@ -640,11 +685,12 @@ class TestThread < Test::Unit::TestCase
   def test_handle_interrupted?
     q = Queue.new
     Thread.handle_interrupt(RuntimeError => :never){
+      done = false
       th = Thread.new{
         q.push :e
         begin
           begin
-            sleep 0.5
+            Thread.pass until done
           rescue => e
             q.push :ng1
           end
@@ -661,6 +707,7 @@ class TestThread < Test::Unit::TestCase
       }
       q.pop
       th.raise
+      done = true
       th.join
       assert_equal(:ok, q.pop)
     }
@@ -706,7 +753,7 @@ _eom
     bug5757 = '[ruby-dev:44985]'
     t0 = Time.now.to_f
     pid = nil
-    cmd = 'r,=IO.pipe; Thread.start {Thread.pass until Thread.main.stop?; puts; STDOUT.flush}; r.read'
+    cmd = 'Signal.trap(:INT, "DEFAULT"); r,=IO.pipe; Thread.start {Thread.pass until Thread.main.stop?; puts; STDOUT.flush}; r.read'
     opt = {}
     opt[:new_pgroup] = true if /mswin|mingw/ =~ RUBY_PLATFORM
     s, _err = EnvUtil.invoke_ruby(['-e', cmd], "", true, true, opt) do |in_p, out_p, err_p, cpid|
@@ -726,8 +773,10 @@ _eom
 
   def test_thread_join_in_trap
     assert_separately [], <<-'EOS'
+    Signal.trap(:INT, "DEFAULT")
+    t0 = Thread.current
     assert_nothing_raised{
-      t = Thread.new{ sleep 0.2; Process.kill(:INT, $$) }
+      t = Thread.new {Thread.pass until t0.stop?; Process.kill(:INT, $$)}
 
       Signal.trap :INT do
         t.join
@@ -736,18 +785,18 @@ _eom
       t.join
     }
     EOS
+  end
 
+  def test_thread_value_in_trap
     assert_separately [], <<-'EOS'
-    assert_equal(:normal_end,
-                 begin
-                   t = Thread.new{ sleep 0.2; Process.kill(:INT, $$); :normal_end }
+    Signal.trap(:INT, "DEFAULT")
+    t0 = Thread.current
+    t = Thread.new {Thread.pass until t0.stop?; Process.kill(:INT, $$); :normal_end}
 
-                   Signal.trap :INT do
-                     t.value
-                   end
-                   t.value
-                 end
-                 )
+    Signal.trap :INT do
+      t.value
+    end
+    assert_equal(:normal_end, t.value)
     EOS
   end
 
@@ -766,14 +815,20 @@ _eom
   end
 
   def test_main_thread_status_at_exit
-    assert_in_out_err([], <<-INPUT, %w(false), [])
+    assert_in_out_err([], <<-'INPUT', ["false false aborting"], [])
+require 'thread'
+q = Queue.new
 Thread.new(Thread.current) {|mth|
   begin
-    sleep 0.1
+    q.push nil
+    mth.run
+    Thread.pass until mth.stop?
+    p :mth_stopped # don't run if killed by rb_thread_terminate_all
   ensure
-    p mth.alive?
+    puts "#{mth.alive?} #{mth.status} #{Thread.current.status}"
   end
 }
+q.pop
     INPUT
   end
 
@@ -783,10 +838,12 @@ Thread.new(Thread.current) {|mth|
     assert_in_out_err([], <<-INPUT, %w(sleep run), [])
       Signal.trap(:INT) {
         puts Thread.current.status
+        exit
       }
+      t = Thread.current
 
       Thread.new(Thread.current) {|mth|
-        sleep 0.01
+        Thread.pass until t.stop?
         puts mth.status
         Process.kill(:INT, $$)
       }
@@ -813,11 +870,13 @@ Thread.new(Thread.current) {|mth|
     }
 
     begin
-      sleep 0.01
+      Thread.pass until ary.size >= 1
+      Thread.pass until t.stop?
       t.kill  # wake up sleep #1
-      sleep 0.01
+      Thread.pass until ary.size >= 2
+      Thread.pass until t.stop?
       t.raise "wakeup" # wake up sleep #2
-      sleep 0.01
+      Thread.pass while t.alive?
       assert_equal(ary, ["run", "aborting", "aborting"])
     ensure
       t.join rescue nil
@@ -844,7 +903,7 @@ Thread.new(Thread.current) {|mth|
         sleep
       }
 
-      sleep 0.01 until th.status == "sleep"
+      Thread.pass until th.status == "sleep"
       # acquired another thread.
       assert_equal(mutex.locked?, true)
       assert_equal(mutex.owned?, false)
@@ -857,15 +916,17 @@ Thread.new(Thread.current) {|mth|
     assert_in_out_err([], <<-INPUT, %w(locked unlocked false), [])
       m = Mutex.new
 
+      trapped = false
       Signal.trap("INT") { |signo|
         m.unlock
+        trapped = true
         puts "unlocked"
       }
 
       m.lock
       puts "locked"
       Process.kill("INT", $$)
-      sleep 0.01
+      Thread.pass until trapped
       puts m.locked?
     INPUT
   end
@@ -891,8 +952,9 @@ Thread.new(Thread.current) {|mth|
                     "0 thread_machine_stack_size")
     assert_operator(h_default[:thread_machine_stack_size], :<=, h_large[:thread_machine_stack_size],
                     "large thread_machine_stack_size")
+  end
 
-    # check VM machine stack size
+  def test_vm_machine_stack_size
     script = 'def rec; print "."; STDOUT.flush; rec; end; rec'
     size_default = invoke_rec script, nil, nil
     assert_operator(size_default, :>, 0, "default size")
@@ -900,9 +962,9 @@ Thread.new(Thread.current) {|mth|
     assert_operator(size_default, :>, size_0, "0 size")
     size_large = invoke_rec script, 1024 * 1024 * 10, nil
     assert_operator(size_default, :<, size_large, "large size")
+  end
 
-    return if /mswin|mingw/ =~ RUBY_PLATFORM
-
+  def test_machine_stack_size
     # check machine stack size
     # Note that machine stack size may not change size (depend on OSs)
     script = 'def rec; print "."; STDOUT.flush; 1.times{1.times{1.times{rec}}}; end; Thread.new{rec}.join'
@@ -912,7 +974,7 @@ Thread.new(Thread.current) {|mth|
     assert_operator(size_default, :>=, size_0, "0 size")
     size_large = invoke_rec script, vm_stack_size, 1024 * 1024 * 10
     assert_operator(size_default, :<=, size_large, "large size")
-  end
+  end unless /mswin|mingw/ =~ RUBY_PLATFORM
 
   def test_blocking_mutex_unlocked_on_fork
     bug8433 = '[ruby-core:55102] [Bug #8433]'
@@ -940,4 +1002,25 @@ Thread.new(Thread.current) {|mth|
     pid, status = Process.waitpid2(pid)
     assert_equal(false, status.success?, bug8433)
   end if Process.respond_to?(:fork)
+
+  def test_fork_in_thread
+    bug9751 = '[ruby-core:62070] [Bug #9751]'
+    f = nil
+    th = Thread.start do
+      unless f = IO.popen("-")
+        STDERR.reopen(STDOUT)
+        exit
+      end
+      Process.wait2(f.pid)
+    end
+    unless th.join(3)
+      Process.kill(:QUIT, f.pid)
+      Process.kill(:KILL, f.pid) unless th.join(1)
+    end
+    _, status = th.value
+    output = f.read
+    f.close
+    assert_not_predicate(status, :signaled?, FailDesc[status, bug9751, output])
+    assert_predicate(status, :success?, bug9751)
+  end if Process.respond_to?(:fork)
 end
