@@ -1,11 +1,11 @@
-#define RUBY_VERSION "2.1.0"
-#define RUBY_RELEASE_DATE "2013-11-23"
+#define RUBY_VERSION "2.2.0"
+#define RUBY_RELEASE_DATE "2014-11-28"
 #define RUBY_PATCHLEVEL -1
 #define RUBY_BRANCH_NAME "trunk"
 
-#define RUBY_RELEASE_YEAR 2013
+#define RUBY_RELEASE_YEAR 2014
 #define RUBY_RELEASE_MONTH 11
-#define RUBY_RELEASE_DAY 23
+#define RUBY_RELEASE_DAY 28
 
 #include "ruby/version.h"
 
@@ -18,7 +18,7 @@
 #endif
 
 #if RUBY_PATCHLEVEL == -1
-#define RUBY_PATCHLEVEL_STR "dev"
+#define RUBY_PATCHLEVEL_STR "preview2"
 #else
 #define RUBY_PATCHLEVEL_STR "p"STRINGIZE(RUBY_PATCHLEVEL)
 #endif
